from typing import Generator, List, TypeVar
from queue import Queue
import threading


T = TypeVar("T")


def batch_generator(gen_in: Generator[T, None, None], batch_size: int) -> Generator[List[T], None, None]:
    """Batches generator, while handling StopIteration

    :param gen_in: generator that generates chunk requests
    :type gen_in: Generator
    """
    batch = []
    for item in gen_in:
        batch.append(item)
        if len(batch) == batch_size:
            yield batch
            batch = []
    if len(batch) > 0:
        yield batch


def prefetch_generator(gen_in: Generator[T, None, None], buffer_size: int) -> Generator[T, None, None]:
    """
    Prefetches from generator while handing StopIteration to ensure items yield immediately.
    Start a thread to prefetch items from the generator and put them in a queue. Upon StopIteration,
    the thread will add a sentinel value to the queue.

    :param gen_in: generator that generates chunk requests
    :type gen_in: Generator
    :param buffer_size: maximum size of the buffer to temporarily store the generators
    :type buffer_size: int
    """
    sentinel = object()
    queue = Queue(maxsize=buffer_size)

    def prefetch():
        for item in gen_in:
            queue.put(item)
        queue.put(sentinel)

    thread = threading.Thread(target=prefetch, daemon=True)
    thread.start()

    while True:
        item = queue.get()
        if item is sentinel:
            break
        yield item


def tail_generator(gen_in: Generator[T, None, None], out_list: List[T]) -> Generator[T, None, None]:
    """Tails generator while handling StopIteration

    :param gen_in: generator that generates chunk requests
    :type gen_in: Generator
    :param out_list: list of tail generators
    :type out_list: List
    """
    for item in gen_in:
        out_list.append(item)
        yield item



import os
import signal
import time
import traceback
from dataclasses import dataclass
from typing import Dict, Any, Optional, List

import typer
from rich.progress import Progress, TextColumn, SpinnerColumn

import skyplane
from skyplane.api.config import TransferConfig, AWSConfig, GCPConfig, AzureConfig, IBMCloudConfig, SCPConfig
from skyplane.api.transfer_job import CopyJob, SyncJob, TransferJob
from skyplane.cli.impl.cp_replicate_fallback import (
    replicate_onprem_cp_cmd,
    replicate_onprem_sync_cmd,
    replicate_small_cp_cmd,
    replicate_small_sync_cmd,
)
from skyplane.cli.impl.common import print_header, console, print_stats_completed, register_exception_handler
from skyplane.api.usage import UsageClient
from skyplane.config import SkyplaneConfig
from skyplane.config_paths import cloud_config, config_path
from skyplane.obj_store.object_store_interface import StorageInterface
from skyplane.cli.impl.progress_bar import ProgressBarTransferHook
from skyplane.utils import logger
from skyplane.utils.definitions import GB, format_bytes
from skyplane.utils.path import parse_path


@dataclass
class TransferStats:
    monitor_status: str
    total_runtime_s: Optional[float] = None
    throughput_gbits: Optional[float] = None
    errors: Optional[Dict[str, List[str]]] = None

    @classmethod
    def empty(cls):
        return TransferStats(monitor_status="empty")

    def to_dict(self) -> Dict[str, Optional[Any]]:
        return {
            "monitor_status": self.monitor_status,
            "total_runtime_s": self.total_runtime_s,
            "throughput_gbits": self.throughput_gbits,
            "errors": [str(e) for e in self.errors.values()] if self.errors else None,
        }


class SkyplaneCLI:
    def __init__(self, src_region_tag: str, dst_region_tag: str, args: Dict[str, Any], skyplane_config: Optional[SkyplaneConfig] = None):
        self.src_region_tag, self.dst_region_tag = src_region_tag, dst_region_tag
        self.args = args
        self.aws_config, self.azure_config, self.gcp_config, self.ibmcloud_config, self.scp_config = self.to_api_config(
            skyplane_config or cloud_config
        )

        # update config
        # TODO: set remaining config params
        if skyplane_config:
            skyplane_config.set_flag("multipart_enabled", str(self.args["multipart"]))
        if cloud_config:
            cloud_config.set_flag("multipart_enabled", str(self.args["multipart"]))

        self.transfer_config = self.make_transfer_config(skyplane_config or cloud_config)
        self.client = skyplane.SkyplaneClient(
            aws_config=self.aws_config,
            azure_config=self.azure_config,
            gcp_config=self.gcp_config,
            transfer_config=self.transfer_config,
            ibmcloud_config=self.ibmcloud_config,
            scp_config=self.scp_config,
        )
        typer.secho(f"Using Skyplane version {skyplane.__version__}", fg="bright_black")

    def to_api_config(self, config: SkyplaneConfig):
        aws_config = AWSConfig(aws_enabled=config.aws_enabled)
        # todo: fix azure config support by collecting azure umi name and resource group and store in skyplane config
        gcp_config = GCPConfig(gcp_project_id=config.gcp_project_id, gcp_enabled=config.gcp_enabled)
        ibmcloud_config = IBMCloudConfig(
            ibmcloud_access_id=config.ibmcloud_access_id,
            ibmcloud_secret_key=config.ibmcloud_secret_key,
            ibmcloud_iam_key=config.ibmcloud_iam_key,
            ibmcloud_iam_endpoint=config.ibmcloud_iam_endpoint,
            ibmcloud_useragent=config.ibmcloud_useragent,
            ibmcloud_resource_group_id=config.ibmcloud_resource_group_id,
            ibmcloud_enabled=config.ibmcloud_enabled,
        )
        scp_config = SCPConfig(
            scp_access_key=config.scp_access_key,
            scp_secret_key=config.scp_secret_key,
            scp_project_id=config.scp_project_id,
            scp_enabled=config.scp_enabled,
        )
        if not config.azure_resource_group or not config.azure_umi_name:
            typer.secho(
                "Azure resource group and UMI name not configured correctly. Please reinit Azure with `skyplane init --reinit-azure`.",
                fg=typer.colors.RED,
                err=True,
            )
            return aws_config, None, gcp_config, ibmcloud_config, scp_config
        azure_config = AzureConfig(
            config.azure_subscription_id,
            config.azure_resource_group,
            config.azure_principal_id,
            config.azure_umi_name,
            config.azure_client_id,
            config.azure_enabled,
        )
        return aws_config, azure_config, gcp_config, ibmcloud_config, scp_config

    def make_transfer_config(self, config: SkyplaneConfig) -> TransferConfig:
        intraregion = self.src_region_tag == self.dst_region_tag
        return TransferConfig(
            autoterminate_minutes=config.get_flag("autoshutdown_minutes"),
            requester_pays=config.get_flag("requester_pays"),
            use_bbr=config.get_flag("bbr"),
            use_compression=config.get_flag("compress") if not intraregion else False,
            use_e2ee=config.get_flag("encrypt_e2e") if not intraregion else False,
            use_socket_tls=config.get_flag("encrypt_socket_tls") if not intraregion else False,
            aws_use_spot_instances=config.get_flag("aws_use_spot_instances"),
            azure_use_spot_instances=config.get_flag("azure_use_spot_instances"),
            gcp_use_spot_instances=config.get_flag("gcp_use_spot_instances"),
            aws_instance_class=config.get_flag("aws_instance_class"),
            azure_instance_class=config.get_flag("azure_instance_class"),
            gcp_instance_class=config.get_flag("gcp_instance_class"),
            ibmcloud_instance_class=config.get_flag("ibmcloud_instance_class"),
            gcp_use_premium_network=config.get_flag("gcp_use_premium_network"),
            scp_instance_class=config.get_flag("scp_instance_class"),
            multipart_enabled=config.get_flag("multipart_enabled"),
            multipart_threshold_mb=config.get_flag("multipart_min_threshold_mb"),
            multipart_chunk_size_mb=config.get_flag("multipart_chunk_size_mb"),
            multipart_max_chunks=config.get_flag("multipart_max_chunks"),
        )

    def check_config(self) -> bool:
        try:
            cloud_config.check_config()
            return True
        except skyplane.exceptions.BadConfigException as e:
            logger.exception(e)
            UsageClient.log_exception("cli_check_config", e, self.args, self.src_region_tag, [self.dst_region_tag])
            return False

    def transfer_cp_onprem(self, src: str, dst: str, recursive: bool) -> bool:
        cmd = replicate_onprem_cp_cmd(src, dst, recursive)
        if cmd:
            typer.secho(f"Delegating to: {cmd}", fg="yellow")
            start = time.perf_counter()
            rc = os.system(cmd)
            request_time = time.perf_counter() - start
            if rc == 0:
                print_stats_completed(request_time, None)
                transfer_stats = TransferStats(monitor_status="completed", total_runtime_s=request_time, throughput_gbits=0)
                UsageClient.log_transfer(transfer_stats.to_dict(), self.args, self.src_region_tag, [self.dst_region_tag])
            return True
        else:
            typer.secho("Transfer not supported", fg="red")
            return True

    def transfer_sync_onprem(self, src: str, dst: str) -> bool:
        cmd = replicate_onprem_sync_cmd(src, dst)
        if cmd:
            typer.secho(f"Delegating to: {cmd}", fg="yellow")
            start = time.perf_counter()
            rc = os.system(cmd)
            request_time = time.perf_counter() - start
            if rc == 0:
                print_stats_completed(request_time, None)
                transfer_stats = TransferStats(monitor_status="completed", total_runtime_s=request_time, throughput_gbits=0)
                UsageClient.log_transfer(transfer_stats.to_dict(), self.args, self.src_region_tag, [self.dst_region_tag])
            return True
        else:
            typer.secho("Transfer not supported", fg="red")
            return True

    def transfer_cp_small(self, src: str, dst: str, recursive: bool) -> bool:
        small_transfer_cmd = replicate_small_cp_cmd(src, dst, recursive)
        if small_transfer_cmd:
            typer.secho(f"Transfer is small enough to delegate to native tools. Delegating to: {small_transfer_cmd}", fg="yellow")
            typer.secho(f"You can disable this with `skyplane config set native_cmd_enabled false`", fg="bright_black")
            os.system(small_transfer_cmd)
            return True
        else:
            return False

    def transfer_sync_small(self, src: str, dst: str) -> bool:
        small_transfer_cmd = replicate_small_sync_cmd(src, dst)
        if small_transfer_cmd:
            typer.secho(f"Transfer is small enough to delegate to native tools. Delegating to: {small_transfer_cmd}", fg="yellow")
            typer.secho(f"You can disable this with `skyplane config set native_cmd_enabled false`", fg="bright_black")
            os.system(small_transfer_cmd)
            return True
        else:
            return False

    # def make_dataplane(self, **solver_args) -> skyplane.Dataplane:
    #    if self.src_region_tag.split(":")[0] == "hdfs":
    #        self.src_region_tag = self.dst_region_tag
    #    dp = self.client.dataplane(*self.src_region_tag.split(":"), *self.dst_region_tag.split(":"), **solver_args)
    #    logger.fs.debug(f"Using dataplane: {dp}")
    #    return dp

    def make_pipeline(self, **solver_args) -> skyplane.Pipeline:
        pipeline = self.client.pipeline(**solver_args)
        logger.fs.debug(f"Using pipeline: {pipeline}")
        return pipeline

    def confirm_transfer(
        self, pipeline: skyplane.Pipeline, src_region_tag: str, dest_region_tags: List[str], query_n: int = 5, ask_to_confirm_transfer=True
    ) -> bool:
        """Prompts the user to confirm their transfer by querying the first query_n files from the TransferJob"""
        if not len(pipeline.jobs_to_dispatch) > 0:
            typer.secho("No jobs to dispatch.")
            return False
        transfer_pair_gen = pipeline.jobs_to_dispatch[0].gen_transfer_pairs()  # type: ignore
        if len(dest_region_tags) == 1:
            console.print(f"[bold yellow]Will transfer objects from {src_region_tag} to {dest_region_tags[0]}[/bold yellow]")
        else:
            console.print(f"[bold yellow]Will transfer objects from {src_region_tag} to {dest_region_tags}[/bold yellow]")

        if src_region_tag.startswith("local") or dest_region_tags[0].startswith("local"):
            # TODO: should still pass cost estimate
            console.print(f"[yellow]Note: local transfers are not monitored by Skyplane[yellow]")
            return True

        topology = pipeline.planner.plan(pipeline.jobs_to_dispatch)
        sorted_counts = sorted(topology.per_region_count().items(), key=lambda x: x[0])
        console.print(
            f"  [bold][blue]VMs to provision:[/blue][/bold] [bright_black]{', '.join(f'{c}x {r}' for r, c in sorted_counts)}[/bright_black]"
        )
        if topology.cost_per_gb:
            console.print(
                f"  [bold][blue]Estimated egress cost:[/blue][/bold] [bright_black]${topology.cost_per_gb:,.2f}/GB[/bright_black]"
            )
        # show spinner
        with Progress(
            TextColumn(" "),
            SpinnerColumn(),
            TextColumn(f"[bright_black]Querying objects for transfer...[/bright_black]"),
            transient=True,
        ) as progress:
            progress.add_task("", total=None)
            obj_pairs = []
            for _ in range(query_n + 1):
                try:
                    obj_pairs.append(next(transfer_pair_gen))
                except StopIteration:
                    break
        if len(obj_pairs) == 0:
            typer.secho("No objects to transfer.")
            return False
        for pair in obj_pairs[:query_n]:
            src_obj = pair.src_obj
            for dst_obj in pair.dst_objs.values():
                console.print(
                    f"  [bright_black][bold]{src_obj.full_path()}[/bold] => [bold]{dst_obj.full_path()}[/bold] ({format_bytes(src_obj.size)})[/bright_black]"
                )
        if len(obj_pairs) > query_n:
            console.print(f"  [bright_black]...[/bright_black]")
        if ask_to_confirm_transfer:
            if typer.confirm("Continue?", default=True):
                logger.fs.debug("User confirmed transfer")
                console.print(
                    "[green]Transfer starting[/green] (Tip: Enable auto-confirmation with `skyplane config set autoconfirm true`)"
                )
                return True
            else:
                logger.fs.error("Transfer cancelled by user.")
                console.print("[bold][red]Transfer cancelled by user.[/red][/bold]")
                raise typer.Abort()
        else:
            console.print("[green]Transfer starting[/green]")
            return True

    def estimate_small_transfer(self, job: TransferJob, size_threshold_bytes: float, query_n: int = 1000) -> bool:
        """Estimates if the transfer is small by querying up to `query_n` files from the TransferJob. If it exceeds
        the file size limit, then it will fall back to the cloud CLIs."""

        # TODO: why shouldn't this include sync?
        if not isinstance(job, CopyJob):
            return False
        transfer_pair_gen = job.gen_transfer_pairs()
        total_size = 0
        generator_exhausted = False
        for _ in range(query_n):
            try:
                pair = next(transfer_pair_gen)
                src_obj = pair.src_obj
                total_size += src_obj.size
                if total_size > size_threshold_bytes:
                    return False
            except StopIteration:
                generator_exhausted = True
                break
        if generator_exhausted and total_size < size_threshold_bytes:
            return True
        else:
            return False


def force_deprovision(dp: skyplane.Dataplane):
    s = signal.signal(signal.SIGINT, signal.SIG_IGN)
    dp.deprovision()
    signal.signal(signal.SIGINT, s)


def run_transfer(
    src: str,
    dst: str,
    recursive: bool,
    debug: bool,
    multipart: bool,
    confirm: bool,
    max_instances: int,
    max_connections: int,
    solver: str,
    cmd: str,
):
    assert cmd == "cp" or cmd == "sync", f"Invalid command: {cmd}"
    if not debug:
        register_exception_handler()
    print_header()

    provider_src, bucket_src, path_src = parse_path(src)
    provider_dst, bucket_dst, path_dst = parse_path(dst)

    # update planner for one-sided transfer
    # somet process for other cloud providers with no VM support
    assert provider_src != "cloudflare" or provider_dst != "cloudflare", "Cannot transfer between two Cloudflare buckets"
    if provider_src == "cloudflare":
        solver = "dst_one_sided"
    elif provider_dst == "cloudflare":
        solver = "src_one_sided"

    src_region_tag = StorageInterface.create(f"{provider_src}:infer", bucket_src).region_tag()
    dst_region_tag = StorageInterface.create(f"{provider_dst}:infer", bucket_dst).region_tag()
    args = {
        "cmd": cmd,
        "recursive": True,
        "debug": debug,
        "multipart": multipart,
        "confirm": confirm,
        "max_instances": max_instances,
        "max_connections": max_connections,
        "solver": solver,
    }

    # create CLI object
    cli = SkyplaneCLI(src_region_tag=src_region_tag, dst_region_tag=dst_region_tag, args=args)
    if not cli.check_config():
        typer.secho(
            f"Skyplane configuration file is not valid. Please reset your config by running `rm {config_path}` and then rerunning `skyplane init` to fix.",
            fg="red",
        )
        return 1

    # create pipeline and queue transfer
    pipeline = cli.make_pipeline(planning_algorithm=solver, max_instances=max_instances)
    if cli.args["cmd"] == "cp":
        pipeline.queue_copy(src, dst, recursive=recursive)
    else:
        pipeline.queue_sync(src, dst)

    # confirm transfer
    if not cli.confirm_transfer(pipeline, src_region_tag, [dst_region_tag], 5, ask_to_confirm_transfer=not confirm):
        return 1

    # local->local transfers not supported (yet)
    if provider_src == "local" and provider_dst == "local":
        raise NotImplementedError("Local->local transfers not supported (yet)")

    # fall back options: local->cloud, cloud->local, small cloud->cloud transfers
    if provider_src == "local" or provider_dst == "local":
        if cli.args["cmd"] == "cp":
            return 0 if cli.transfer_cp_onprem(src, dst, recursive) else 1
        else:
            return 0 if cli.transfer_sync_onprem(src, dst) else 1
    elif cloud_config.get_flag("native_cmd_enabled"):
        # fallback option: transfer is too small
        if cli.args["cmd"] == "cp":
            job = CopyJob(src, [dst], recursive=recursive)  # TODO: rever to using pipeline
            if cli.estimate_small_transfer(job, cloud_config.get_flag("native_cmd_threshold_gb") * GB):
                small_transfer_status = cli.transfer_cp_small(src, dst, recursive)
                return 0 if small_transfer_status else 1
        else:
            job = SyncJob(src, [dst])
            if cli.estimate_small_transfer(job, cloud_config.get_flag("native_cmd_threshold_gb") * GB):
                small_transfer_status = cli.transfer_sync_small(src, dst)
                return 0 if small_transfer_status else 1

    # dataplane must be created after transfers are queued
    dp = pipeline.create_dataplane(debug=debug)
    with dp.auto_deprovision():
        try:
            dp.provision(spinner=True)
            dp.run(pipeline.jobs_to_dispatch, hooks=ProgressBarTransferHook(dp.topology.dest_region_tags))
        except KeyboardInterrupt:
            logger.fs.warning("Transfer cancelled by user (KeyboardInterrupt).")
            console.print("\n[red]Transfer cancelled by user. Copying gateway logs and exiting.[/red]")
            try:
                dp.copy_gateway_logs()
                force_deprovision(dp)
            except Exception as e:
                logger.fs.exception(e)
                console.print(f"[bright_black]{traceback.format_exc()}[/bright_black]")
                console.print(e)
                UsageClient.log_exception("cli_cp", e, args, cli.src_region_tag, [cli.dst_region_tag])
                console.print("[bold red]Deprovisioning was interrupted! VMs may still be running which will incur charges.[/bold red]")
                console.print("[bold red]Please manually deprovision the VMs by running `skyplane deprovision`.[/bold red]")
            return 1
        except skyplane.exceptions.SkyplaneException as e:
            console.print(f"[bright_black]{traceback.format_exc()}[/bright_black]")
            console.print(e.pretty_print_str())
            UsageClient.log_exception("cli_query_objstore", e, args, cli.src_region_tag, [cli.dst_region_tag])
            force_deprovision(dp)
        except Exception as e:
            logger.fs.exception(e)
            console.print(f"[bright_black]{traceback.format_exc()}[/bright_black]")
            console.print(e)
            UsageClient.log_exception("cli_query_objstore", e, args, cli.src_region_tag, [cli.dst_region_tag])
            force_deprovision(dp)


def cp(
    src: str,
    dst: str,
    recursive: bool = typer.Option(False, "--recursive", "-r", help="If true, will copy objects at folder prefix recursively"),
    debug: bool = typer.Option(False, help="If true, will write debug information to debug directory."),
    multipart: bool = typer.Option(cloud_config.get_flag("multipart_enabled"), help="If true, will use multipart uploads."),
    # transfer flags
    confirm: bool = typer.Option(cloud_config.get_flag("autoconfirm"), "--confirm", "-y", "-f", help="Confirm all transfer prompts"),
    max_instances: int = typer.Option(cloud_config.get_flag("max_instances"), "--max-instances", "-n", help="Number of gateways"),
    max_connections: int = typer.Option(
        cloud_config.get_flag("num_connections"), "--max-connections", help="Number of connections per gateway"
    ),
    # todo - add solver params once API supports it
    # solver
    solver: str = typer.Option("direct", "--solver", help="Solver to use for transfer"),
    solver_required_throughput_gbits: float = typer.Option(1, "--tput", "-t", help="Required throughput to be solved for in Gbps"),
):
    """
    `cp` copies a file or folder from one location to another. If the source is on an object store,
    it will copy all objects with that prefix. If it is a local path, it will copy the entire file
    or directory tree.

    By default, it will copy objects using a direct connection between instances. However, if you would
    like to use the solver, call `--solve`. Note that the solver requires a throughput grid file to be
    specified. We provide a default one but it may be out-of-date.

    :param src: Source prefix to copy from
    :type src: str
    :param dst: The destination of the transfer
    :type dst: str
    :param recursive: If true, will copy objects at folder prefix recursively
    :type recursive: bool
    :param debug: If true, will write debug information to debug directory.
    :type debug: bool
    :param multipart: If true, will use multipart uploads.
    :type multipart: bool
    :param confirm: If true, will not prompt for confirmation of transfer.
    :type confirm: bool
    :param max_instances: The maximum number of instances to use per region (default: 1)
    :type max_instances: int
    :param max_connections: The maximum number of connections per instance (default: 32)
    :type max_connections: int
    :param solver: The solver to use for the transfer (default: direct)
    :type solver: str
    """
    return run_transfer(src, dst, recursive, debug, multipart, confirm, max_instances, max_connections, solver, "cp")


def sync(
    src: str,
    dst: str,
    debug: bool = typer.Option(False, help="If true, will write debug information to debug directory."),
    multipart: bool = typer.Option(cloud_config.get_flag("multipart_enabled"), help="If true, will use multipart uploads."),
    # transfer flags
    confirm: bool = typer.Option(cloud_config.get_flag("autoconfirm"), "--confirm", "-y", "-f", help="Confirm all transfer prompts"),
    max_instances: int = typer.Option(cloud_config.get_flag("max_instances"), "--max-instances", "-n", help="Number of gateways"),
    max_connections: int = typer.Option(
        cloud_config.get_flag("num_connections"), "--max-connections", help="Number of connections per gateway"
    ),
    # todo - add solver params once API supports it
    # solver
    solver: str = typer.Option("direct", "--solver", help="Solver to use for transfer"),
    solver_required_throughput_gbits: float = typer.Option(1, "--tput", "-t", help="Required throughput to be solved for"),
):
    """
    'sync` synchronizes files or folders from one location to another. If the source is on an object store,
    it will copy all objects with that prefix. If it is a local path, it will copy the entire file
    or directory tree.

    By default, it will copy objects using a direct connection between instances. However, if you would
    like to use the solver, call `--solve`. Note that the solver requires a throughput grid file to be
    specified. We provide a default one but it may be out-of-date.

    For each file in the source, it is copied over if the file does not exist in the destination, it has
    a different size in the destination, or if the source version of the file was more recently modified
    than the destination. This behavior is similar to 'aws sync'.

    :param src: Source prefix to copy from
    :type src: str
    :param dst: The destination of the transfer
    :type dst: str
    :param debug: If true, will write debug information to debug directory.
    :type debug: bool
    :param multipart: If true, will use multipart uploads.
    :type multipart: bool
    :param confirm: If true, will not prompt for confirmation of transfer.
    :type confirm: bool
    :param max_instances: The maximum number of instances to use per region (default: 1)
    :type max_instances: int
    :param max_connections: The maximum number of connections per instance (default: 32)
    :type max_connections: int
    :param solver: The solver to use for the transfer (default: direct)
    :type solver: str
    """
    return run_transfer(src, dst, False, debug, multipart, confirm, max_instances, max_connections, solver, "sync")



# (C) Copyright Samsung SDS. 2023

#

# Licensed under the Apache License, Version 2.0 (the "License");

# you may not use this file except in compliance with the License.

# You may obtain a copy of the License at

#

#     http://www.apache.org/licenses/LICENSE-2.0

#

# Unless required by applicable law or agreed to in writing, software

# distributed under the License is distributed on an "AS IS" BASIS,

# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

# See the License for the specific language governing permissions and

# limitations under the License.
import os
from typing import List, Optional
import uuid
from pathlib import Path

from skyplane.compute.key_utils import generate_keypair

from skyplane import exceptions
from skyplane.compute.scp.scp_auth import SCPAuthentication
from skyplane.compute.scp.scp_server import SCPServer
from skyplane.compute.scp.scp_network import SCPNetwork
from skyplane.compute.cloud_provider import CloudProvider
from skyplane.compute.server import key_root
from skyplane.utils import logger
from skyplane.utils.fn import do_parallel


class SCPCloudProvider(CloudProvider):
    def __init__(
        self,
        key_prefix: str = "skyplane",
        key_root=key_root / "scp",
        auth: Optional[SCPAuthentication] = None,
        network: Optional[SCPNetwork] = None,
    ):
        super().__init__()
        self.key_name = key_prefix
        self.auth = auth if auth else SCPAuthentication()
        self.network = network if network else SCPNetwork(self.auth)
        key_root.mkdir(parents=True, exist_ok=True)
        self.private_key_path = key_root / "scp_key"
        self.public_key_path = key_root / "scp_key.pub"

    @property
    def name(self):
        return "scp"

    @staticmethod
    def region_list() -> List[str]:
        return SCPAuthentication.get_region_config()

    @classmethod
    def get_transfer_cost(cls, src_key, dst_key, premium_tier=True):
        assert src_key.startswith("scp:")
        dst_provider, dst_region = dst_key.split(":")

        return 0.077

    def create_ssh_key(self):
        public_key_path = Path(self.public_key_path)
        private_key_path = Path(self.private_key_path)
        if not private_key_path.exists():
            private_key_path.parent.mkdir(parents=True, exist_ok=True)
            generate_keypair(public_key_path, private_key_path)

    def get_init_script(self):
        cmd_st = "mkdir -p ~/.ssh/; touch ~/.ssh/authorized_keys;"
        with open(os.path.expanduser(self.public_key_path)) as f:
            pub_key = f.read()
        cmd = "echo '{}' &>>~/.ssh/authorized_keys;".format(pub_key)
        cmd_ed = "chmod 644 ~/.ssh/authorized_keys; chmod 700 ~/.ssh/; "
        cmd_ed += "mkdir -p ~/.scp; "

        return cmd_st + cmd + cmd_ed

    def get_instance_list(self, region) -> List[SCPServer]:
        try:
            service_zone_id = self.network.get_service_zone_id(region)
            servergroups = self.network.list_server_group()
            servergroups = [sg for sg in servergroups if sg["serviceZoneId"] == service_zone_id]

            instances = []
            for sg in servergroups:
                instances += self.network.list_server_group_instance(sg["serverGroupId"])

            instancelist = []
            validstate = ["RUNNING", "STOPPED", "STOPPING", "STARTING"]
            for instance in instances:
                if instance["virtualServerState"] in validstate:
                    instancelist.append(
                        {"virtualServerId": instance["virtualServerId"], "virtualServerName": instance["virtualServerName"]}
                    )

            return [
                SCPServer(f"scp:{region}", virtualServerName=i["virtualServerName"], virtualServerId=i["virtualServerId"])
                for i in instancelist
            ]
        except Exception as e:
            raise exceptions.SkyplaneException(f"Failed to get instance list: {e}")

    def setup_global(self, **kwargs):
        pass

    def setup_region(self, region: str):
        self.network.make_vpc(region)

    def provision_instance(
        self,
        region: str,
        instance_class: str,
        disk_size: int = 100,
        use_spot_instances: bool = False,
        name: Optional[str] = None,
        tags={"skyplane": "true"},
        instance_os: str = "ubuntu",
    ) -> SCPServer:
        assert not region.startswith("scp:"), "Region should be SCP region"
        if name is None:
            name = f"skyplane-scp-{uuid.uuid4().hex[:8]}"
        url = f"/virtual-server/v3/virtual-servers"

        service_zone_id = self.network.get_service_zone_id(region)

        vpc_name = self.network.define_vpc_name(region)
        vpc_id = self.network.list_vpcs(service_zone_id, vpc_name)[0]["vpcId"]
        init_script = self.get_init_script()
        subnet_id = self.network.list_subnets(vpc_id, None)[0]["subnetId"]
        security_group_id = self.network.list_security_groups(vpc_id, None)[0]["securityGroupId"]
        servergroup = tags.get("servergroup", "default")
        image_id = "IMAGE-mX_5UKOJqriGnWZ2GXEbEg"  # Ubuntu 22.04

        req_body = {
            "availabilityZoneName": "AZ1" if region == "KOREA-WEST-MAZ-SCP-B001" else None,
            "blockStorage": {
                "blockStorageName": "Skyplane-Block-Storage",
                "diskSize": disk_size,
            },
            "imageId": image_id,
            "initialScript": {
                "encodingType": "plain",
                "initialScriptShell": "bash",
                "initialScriptType": "text",
                "initialScriptContent": init_script,
            },
            "nic": {
                "natEnabled": "true",
                # "publicIpId" : "PUBLIC_IP-xxxxxx",
                "subnetId": subnet_id,
            },
            "osAdmin": {"osUserId": "root", "osUserPassword": "test123$"},
            "securityGroupIds": [security_group_id],
            "serverGroupId": servergroup,
            "serverType": instance_class,
            "serviceZoneId": service_zone_id,
            "tags": [
                {"tagKey": "skyplane", "tagValue": "true"},
                {"tagKey": "skyplaneclientid", "tagValue": tags.get("skyplaneclientid", "unknown")},
            ],
            "virtualServerName": name,
        }
        try:
            # self.network.scp_client.random_wait()
            response = self.network.scp_client._post(url, req_body)
            # print(response)
            virtualServerId = response["resourceId"]

            # wait for server editing
            create_completion_condition = lambda: self.network.get_vs_details(virtualServerId)["virtualServerState"] == "EDITING"
            self.network.scp_client.wait_for_completion(
                f"[{region}:{name}:{instance_class}] Creating SCP Skyplane Virtual Server", create_completion_condition
            )

            # wait for virtual server ip
            create_completion_condition = lambda: self.network.get_vs_details(virtualServerId).get("ip") is not None
            self.network.scp_client.wait_for_completion(
                f"[{region}:{name}:{instance_class}] Creating SCP Skyplane Virtual Server ip", create_completion_condition
            )

            # Add firewall rule
            self.network.add_firewall_22_rule(region, servergroup, virtualServerId)

            # wait for server running
            create_completion_condition = lambda: self.network.get_vs_details(virtualServerId)["virtualServerState"] == "RUNNING"
            self.network.scp_client.wait_for_completion(
                f"[{region}:{name}:{instance_class}] Running SCP Skyplane Virtual Server", create_completion_condition
            )

            server = SCPServer(f"scp:{region}", virtualServerName=name, virtualServerId=virtualServerId, vpc_id=vpc_id)

        except KeyboardInterrupt:
            logger.warning(f"keyboard interrupt. You may need to use the deprovision command to terminate instance {name}.")
            # wait for server running
            create_completion_condition = lambda: self.network.get_vs_details(virtualServerId)["virtualServerState"] == "RUNNING"
            self.network.scp_client.wait_for_completion(
                f"[{region}:{name}:{instance_class}] Running SCP Skyplane Virtual Server", create_completion_condition
            )
            server = SCPServer(f"scp:{region}", virtualServerName=name, virtualServerId=virtualServerId, vpc_id=vpc_id)
            server.terminate_instance_impl()
            raise
        except Exception as e:
            # Not found any cluster by requested Scale
            if "404" in str(e):
                raise Exception(
                    f"Failed to provision the instance : due to insufficient cluster capacity compared to the requested scale, {e}. Check the SCP console for more details."
                )
            else:
                raise Exception(f"Exception occurred during provision instance: {e}. Check the SCP console for more details.")
        return server

    def add_firewall_rule_all(self, region, private_ips, vpcids):
        self.network.add_firewall_rule_all(region, private_ips, vpcids)

    def remove_gateway_rule(self, server: SCPServer):
        ip = server.private_ip()
        igw = self.network.get_igw(server.vpc_id)
        firewall = self.network.get_firewallId(igw)

        ruleIds = self.network.list_firewall_rules(firewall)

        ruleDetails = [self.network.get_firewall_rule_details(firewall, id) for id in ruleIds]
        # for d in ruleDetails: print(d)
        ruleDetails = [detail for detail in ruleDetails if "ruleId" in detail]  # HACK, some times wrong responses

        rule_source = [detail["ruleId"] for detail in ruleDetails if ip in detail["sourceIpAddresses"]]
        rule_dest = [detail["ruleId"] for detail in ruleDetails if ip in detail["destinationIpAddresses"]]
        rule_ids = list(set(rule_source + rule_dest))
        print(f"remove_gateway_rule : rule_ids - {rule_ids}")
        self.network.delete_firewall_rules(server.region(), firewall, rule_ids)

    def remove_gateway_rule_region(self, region, private_ips, vpcids):
        igw = self.network.get_igw(vpcids[0])
        firewallId = self.network.get_firewallId(igw)

        ruleIds = self.network.list_firewall_rules(firewallId)

        # For each private IP, check if it appears in the source or destination IPs of ruleIds' details.
        ruleDetails = [self.network.get_firewall_rule_details(firewallId, id) for id in ruleIds]

        rule_ids = list(
            set(
                detail["ruleId"]
                for ip in private_ips
                for detail in ruleDetails
                if ip in detail["sourceIpAddresses"] or ip in detail["destinationIpAddresses"]
            )
        )

        self.network.delete_firewall_rules(region, firewallId, rule_ids)

    def remove_gateway_rule_all(self, instances):
        # Get all regions
        regions = list(set([instance.region() for instance in instances]))

        # Get all private IPs and VPC IDs for each region, then remove gateway rules for each region
        for region in regions:
            private_ips = [instance.private_ip() for instance in instances if instance.region() == region]
            vpcids = [instance.vpc_id for instance in instances if instance.region() == region]
            self.remove_gateway_rule_region(region, private_ips, vpcids)

    def teardown_global(self):
        args = []
        for region in self.region_list():
            service_zone_id = self.network.get_service_zone_id(region)
            vpc_name = self.network.define_vpc_name(region)
            skyplane_vpc = self.network.find_vpc_id(service_zone_id, vpc_name)
            if skyplane_vpc:
                args.append((region, skyplane_vpc))

        if len(args) > 0:
            do_parallel(lambda x: self.network.delete_vpc(*x), args, desc="Removing VPCs", spinner=True, spinner_persist=True)



import json
from pathlib import Path
import os
from typing import List
import queue
import socket
import ssl
import time
import lz4.frame
import traceback
from functools import partial
from multiprocessing import Event, Process, Queue
from multiprocessing.managers import DictProxy
from typing import Dict, List, Optional

import urllib3
import nacl.secret
from abc import ABC, abstractmethod

from skyplane.config_paths import cloud_config
from skyplane.utils.definitions import MB
from skyplane.utils import logger
from skyplane.utils.retry import retry_backoff
from skyplane.utils.timer import Timer
from skyplane.obj_store.object_store_interface import ObjectStoreInterface

from skyplane.chunk import ChunkRequest, ChunkState
from skyplane.gateway.gateway_queue import GatewayQueue
from skyplane.gateway.chunk_store import ChunkStore


class GatewayOperator(ABC):
    def __init__(
        self,
        handle: str,
        region: str,  # TODO: remove
        input_queue: GatewayQueue,
        output_queue: GatewayQueue,
        error_event,
        error_queue: Queue,
        chunk_store: ChunkStore,
        n_processes: Optional[int] = 1,
    ):
        self.handle = handle
        self.region = region
        self.input_queue = input_queue
        self.output_queue = output_queue
        self.chunk_store = chunk_store
        self.error_event = error_event
        self.error_queue = error_queue
        self.n_processes = n_processes

        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        # args for worker function
        self.args = ()

        # shared state
        # TODO: move this outside of GateWaySender to share between processes?
        self.processes = []
        self.exit_flags = [Event() for _ in range(self.n_processes)]

        # process-local state
        self.worker_id: Optional[int] = None

    def start_workers(self):
        for i in range(self.n_processes):
            p = Process(target=self.worker_loop, args=(i,) + self.args)
            p.start()
            self.processes.append(p)

    def stop_workers(self):
        for i in range(self.n_processes):
            self.exit_flags[i].set()
        for p in self.processes:
            p.join()
        self.processes = []

    def worker_loop(self, worker_id: int, *args):
        self.worker_id = worker_id
        while not self.exit_flags[worker_id].is_set() and not self.error_event.is_set():
            try:
                # get chunk from input queue
                try:
                    # will only get data for that handle
                    chunk_req = self.input_queue.get_nowait(self.handle)
                except queue.Empty:
                    continue

                # TODO: status logging
                self.chunk_store.log_chunk_state(chunk_req, ChunkState.in_progress, operator_handle=self.handle, worker_id=worker_id)
                # process chunk
                succ = self.process(chunk_req, *args)

                # place in output queue
                if succ:
                    self.chunk_store.log_chunk_state(chunk_req, ChunkState.complete, operator_handle=self.handle, worker_id=worker_id)
                    if self.output_queue is not None:
                        self.output_queue.put(chunk_req)
                    else:
                        print(f"[{self.handle}:{self.worker_id}] Output queue is None - terminal operator")
                    time.sleep(0.1)  # yield ?
                else:
                    # failed to process - re-queue
                    time.sleep(0.1)
                    self.input_queue.put(chunk_req)

            except Exception as e:
                logger.error(f"[{self.handle}:{self.worker_id}] Exception: {e}")
                self.error_queue.put(traceback.format_exc())
                self.error_event.set()
                self.exit_flags[worker_id].set()

        # run worker exit function
        self.worker_exit(worker_id)

    def worker_exit(self, worker_id: int):
        pass

    @abstractmethod
    def process(self, chunk_req: ChunkRequest, **args):
        pass


class GatewayWaitReceiver(GatewayOperator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # TODO: alternative (potentially better performnace) implementation: connect via queue with GatewayReceiver to listen
    # for download completition events - join with chunk request queue from ChunkStore
    def process(self, chunk_req: ChunkRequest):
        chunk_file_path = self.chunk_store.get_chunk_file_path(chunk_req.chunk.chunk_id)
        if not os.path.exists(chunk_file_path):  # chunk still not downloaded, re-queue
            # logger.debug(f"[{self.handle}:{self.worker_id}] Chunk {chunk_req.chunk.chunk_id} not downloaded yet, re-queueing")
            return False

        # check to see if file is completed downloading
        # Successfully received chunk 38400a29812142a486eaefcdebedf371, 161867776    0, 67108864
        with open(chunk_file_path, "rb") as f:
            data = f.read()
            if len(data) < chunk_req.chunk.chunk_length_bytes:
                # download not complete
                return False
            assert (
                len(data) == chunk_req.chunk.chunk_length_bytes
            ), f"Downloaded chunk length does not match expected length: {len(data)}, {chunk_req.chunk.chunk_length_bytes}"
        print(
            f"[{self.handle}:{self.worker_id}] Successfully received chunk {chunk_req.chunk.chunk_id}, {len(data)}, {chunk_req.chunk.chunk_length_bytes}"
        )
        return True


class GatewaySender(GatewayOperator):
    def __init__(
        self,
        handle: str,
        region: str,
        input_queue: GatewayQueue,
        output_queue: GatewayQueue,
        error_event,
        error_queue: Queue,
        chunk_store: ChunkStore,
        ip_addr: str,
        use_tls: Optional[bool] = True,
        use_compression: Optional[bool] = True,
        e2ee_key_bytes: Optional[bytes] = None,
        n_processes: Optional[int] = 32,
    ):
        super().__init__(handle, region, input_queue, output_queue, error_event, error_queue, chunk_store, n_processes)
        self.ip_addr = ip_addr
        self.use_tls = use_tls
        self.use_compression = use_compression
        self.e2ee_key_bytes = e2ee_key_bytes
        self.args = (ip_addr,)

        # provider = region.split(":")[0]
        # if provider == "aws" or provider == "gcp":
        #    self.n_processes = 32
        # elif provider == "azure":
        #    self.n_processes = 24  # due to throttling limits from authentication

        # encryption
        if e2ee_key_bytes is None:
            self.e2ee_secretbox = None
        else:
            self.e2ee_secretbox = nacl.secret.SecretBox(e2ee_key_bytes)

        # SSL context
        if use_tls:
            self.ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            self.ssl_context.check_hostname = False
            self.ssl_context.verify_mode = ssl.CERT_NONE
            logger.info(f"Using {str(ssl.OPENSSL_VERSION)}")
        else:
            self.ssl_context = None

        # process-local state
        self.sender_port: Optional[int] = None
        self.destination_ports: Dict[str, int] = {}  # ip_address -> int
        self.destination_sockets: Dict[str, socket.socket] = {}  # ip_address -> socket
        self.sent_chunk_ids: Dict[str, List[int]] = {}  # ip_address -> list of chunk_ids

        # http pool
        timeout = urllib3.util.Timeout(connect=10.0, read=None)  # no read timeout
        self.http_pool = urllib3.PoolManager(retries=urllib3.Retry(total=3), cert_reqs="CERT_NONE", timeout=timeout)

    def worker_exit(self, worker_id: int):
        # close destination sockets
        logger.info(f"[sender:{worker_id}] exiting, closing sockets")
        for dst_socket in self.destination_sockets.values():
            dst_socket.close()

        # wait for all chunks to reach state "downloaded"
        # TODO: remove/replace for broadcast
        def wait_for_chunks():
            cr_status = {}
            for ip, ip_chunk_ids in self.sent_chunk_ids.items():
                response = self.http_pool.request("GET", f"https://{ip}:8080/api/v1/incomplete_chunk_requests")
                assert response.status == 200, f"{response.status} {response.data}"
                host_state = json.loads(response.data.decode("utf-8"))["chunk_requests"]
                for chunk_id in ip_chunk_ids:
                    if chunk_id in host_state:
                        cr_status[chunk_id] = host_state[chunk_id]["state"]
            return all(status not in ["registered", "download_queued", "download_in_progress"] for status in cr_status.values())

        logger.info(f"[sender:{worker_id}] waiting for chunks to reach state 'downloaded'")
        wait_success = False
        for _ in range(60):
            try:
                if wait_for_chunks():
                    wait_success = True
                    break
            except Exception as e:
                logger.error(f"[Gateway Sender wait_for_chunks()] Exception: {e}")

            time.sleep(5)  # originally 1
        if not wait_success:
            raise Exception("Timed out waiting for chunks to reach state 'downloaded'")
        logger.info(f"[sender:{worker_id}] all chunks reached state 'downloaded'")

        # close servers
        logger.info(f"[sender:{worker_id}] exiting, closing servers")
        for dst_host, dst_port in self.destination_ports.items():
            response = self.http_pool.request("DELETE", f"https://{dst_host}:8080/api/v1/servers/{dst_port}")
            assert response.status == 200 and json.loads(response.data.decode("utf-8")) == {"status": "ok"}
            logger.info(f"[sender:{worker_id}] closed destination socket {dst_host}:{dst_port}")

    def make_socket(self, dst_host):
        response = self.http_pool.request("POST", f"https://{dst_host}:8080/api/v1/servers")
        assert response.status == 200, f"{response.status} {response.data.decode('utf-8')}"
        self.destination_ports[dst_host] = int(json.loads(response.data.decode("utf-8"))["server_port"])
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((dst_host, self.destination_ports[dst_host]))
        original_timeout = sock.gettimeout()
        sock.settimeout(30.0)  # For the TLS handshake
        logger.info(f"[sender:{self.worker_id}] started new server connection to {dst_host}:{self.destination_ports[dst_host]}")
        if self.ssl_context is not None:
            sock = self.ssl_context.wrap_socket(sock)
            logger.info(f"[sender:{self.worker_id}] finished TLS handshake to {dst_host}:{self.destination_ports[dst_host]}")
        sock.settimeout(original_timeout)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        return sock

    # send chunks to other instances
    def process(self, chunk_req: ChunkRequest, dst_host: str):
        """Send list of chunks to gateway server, pipelining small chunks together into a single socket stream."""
        # notify server of upcoming ChunkRequests

        # print(f"[{self.handle}:{self.worker_id}] Sending chunk ID {chunk_req.chunk.chunk_id} to IP {dst_host}")

        # TODO: does this function need to be implemented to work for a list of chunks?

        chunk_ids = [chunk_req.chunk.chunk_id]
        chunk_reqs = [chunk_req]

        # retry http_pool.request
        def http_pool_request(register_body):
            return self.http_pool.request(
                "POST",
                f"https://{dst_host}:8080/api/v1/chunk_requests",
                body=register_body,
                headers={"Content-Type": "application/json"},
            )

        try:
            with Timer(f"pre-register chunks {chunk_ids} to {dst_host}"):
                # TODO: remove chunk request wrapper
                # while True:
                #   try:
                #       response = self.http_pool.request(
                #           "POST", f"https://{dst_host}:8080/api/v1/chunk_requests", body=register_body, headers={"Content-Type": "application/json"}
                #       )
                #       break
                #   except Exception as e:
                #       print("sender post error", e)
                #       time.sleep(1)
                n_added = 0
                while n_added < len(chunk_reqs):
                    register_body = json.dumps([c.chunk.as_dict() for c in chunk_reqs[n_added:]]).encode("utf-8")
                    # print(f"[sender-{self.worker_id}]:{chunk_ids} register body {register_body}")
                    # response = self.http_pool.request(
                    #     "POST",
                    #     f"https://{dst_host}:8080/api/v1/chunk_requests",
                    #     body=register_body,
                    #     headers={"Content-Type": "application/json"},
                    # )
                    response = retry_backoff(partial(http_pool_request, register_body))
                    reply_json = json.loads(response.data.decode("utf-8"))
                    print(f"[sender-{self.worker_id}]", n_added, reply_json, dst_host)
                    n_added += reply_json["n_added"]
                    assert response.status == 200, f"Wrong response status {response.status}"
                    # json.loads(response.data.decode("utf-8")).get("status") == "ok"
                    if n_added == len(chunk_reqs):
                        print(f"[sender-{self.worker_id}]:{chunk_ids} registered chunks")
                    else:
                        time.sleep(1)
        except Exception as e:
            print(f"[{self.handle}:{self.worker_id}] Error registering chunks {chunk_ids} to {dst_host}: {e}")
            raise e

        # contact server to set up socket connection
        def socket_connection():
            self.destination_sockets[dst_host] = retry_backoff(
                # all exceptions are retried
                # partial(self.make_socket, dst_host), max_retries=3, exception_class=socket.timeout
                partial(self.make_socket, dst_host),
                max_retries=3,
                exception_class=Exception,
            )

        if self.destination_ports.get(dst_host) is None:
            print(f"[sender-{self.worker_id}]:{chunk_ids} creating new socket")
            # self.destination_sockets[dst_host] = retry_backoff(
            #     # all exceptions are retried
            #     # partial(self.make_socket, dst_host), max_retries=3, exception_class=socket.timeout
            #     partial(self.make_socket, dst_host), max_retries=3, exception_class=Exception
            # )
            socket_connection()
            print(f"[sender-{self.worker_id}]:{chunk_ids} created new socket")
        sock = self.destination_sockets[dst_host]

        # TODO: cleanup so this isn't a loop
        for idx, chunk_req in enumerate(chunk_reqs):
            # self.chunk_store.state_start_upload(chunk_id, f"sender:{self.worker_id}")
            chunk_id = chunk_req.chunk.chunk_id
            chunk = chunk_req.chunk
            chunk_file_path = self.chunk_store.get_chunk_file_path(chunk_id)

            # read data from disk (and optionally compress if sending from source region)
            with open(chunk_file_path, "rb") as f:
                data = f.read()
            assert len(data) == chunk.chunk_length_bytes, f"chunk {chunk_id} has size {len(data)} but should be {chunk.chunk_length_bytes}"

            wire_length = len(data)
            raw_wire_length = wire_length
            compressed_length = None

            if self.use_compression:
                data = lz4.frame.compress(data)
                wire_length = len(data)
                compressed_length = wire_length
            if self.e2ee_secretbox is not None:
                data = self.e2ee_secretbox.encrypt(data)
                wire_length = len(data)

            # send chunk header
            header = chunk.to_wire_header(
                n_chunks_left_on_socket=len(chunk_ids) - idx - 1,
                wire_length=wire_length,
                raw_wire_length=raw_wire_length,
                is_compressed=(compressed_length is not None),
            )
            # print(f"[sender-{self.worker_id}]:{chunk_id} sending chunk header {header}")
            # fix socket timeouterror case - we need to create a new socket, so retry_backoff is not easy to use.
            retry_count = 0
            max_retries = 3
            while retry_count < max_retries:
                try:
                    header.to_socket(sock)
                    # print(f"[sender-{self.worker_id}]:{chunk_id} sent chunk header")

                    # send chunk data
                    assert chunk_file_path.exists(), f"chunk file {chunk_file_path} does not exist"
                    # file_size = os.path.getsize(chunk_file_path)

                    with Timer() as t:
                        sock.sendall(data)
                except Exception as e:
                    print(f"[sender-{self.worker_id}]:{chunk_id} error sending chunk {e}")
                    retry_count += 1
                    if retry_count < max_retries:
                        # print(f"SCP DEV - Retrying (attempt {retry_count} of {max_retries})...")
                        # with open(f"/skyplane/header.to_socket_{retry_count}_{dst_host}_error.txt", "w") as f:
                        #    f.write(str(e))
                        time.sleep(0.5)
                        socket_connection()
                        sock = self.destination_sockets[dst_host]
                    else:
                        # print(f"SCP DEV - Max retries ({max_retries}) reached. Giving up.")
                        raise e
                else:
                    break

            # logger.debug(f"[sender:{self.worker_id}]:{chunk_id} sent at {chunk.chunk_length_bytes * 8 / t.elapsed / MB:.2f}Mbps")
            print(
                f"[sender:{self.worker_id}]:{chunk_id} sent at {wire_length * 8 / t.elapsed / MB:.2f}Mbps - {dst_host} : {self.destination_ports[dst_host]}"
            )

            if dst_host not in self.sent_chunk_ids:
                self.sent_chunk_ids[dst_host] = []
            self.sent_chunk_ids[dst_host].append(chunk_req.chunk.chunk_id)

        # success, so return true
        return True


class GatewayRandomDataGen(GatewayOperator):
    def __init__(
        self,
        handle: str,
        region: str,
        input_queue: GatewayQueue,
        output_queue: GatewayQueue,
        error_event,
        error_queue: Queue,
        chunk_store: ChunkStore,
        size_mb: int,
        n_processes: Optional[int] = 1,
    ):
        super().__init__(handle, region, input_queue, output_queue, error_event, error_queue, chunk_store, n_processes)
        self.size_mb = size_mb

    def process(self, chunk_req: ChunkRequest):
        # wait until enough space available
        fpath = str(self.chunk_store.get_chunk_file_path(chunk_req.chunk.chunk_id).absolute())
        size_bytes = int(self.size_mb * MB)
        assert size_bytes > 0, f"Invalid size {size_bytes} for fallocate"

        while True:
            # create file with random data
            try:
                os.system(f"fallocate -l {size_bytes} {fpath}")
                file_size = os.path.getsize(fpath)
                if file_size == size_bytes:
                    break
            except Exception:
                print(f"[gen_data] Chunk store full, waiting before generating {chunk_req.chunk.chunk_id}")
                time.sleep(0.1)
                continue

        logger.info(f"[{self.handle}:{self.worker_id}] Wrote chunk {chunk_req.chunk.chunk_id} with size {file_size} to {fpath}")
        chunk_req.chunk.chunk_length_bytes = os.path.getsize(fpath)

        return True


class GatewayWriteLocal(GatewayOperator):
    def __init__(
        self,
        handle: str,
        region: str,
        input_queue: GatewayQueue,
        output_queue: GatewayQueue,
        error_event,
        error_queue: Queue,
        chunk_store: ChunkStore,
        n_processes: int = 1,
    ):
        super().__init__(handle, region, input_queue, output_queue, error_event, error_queue, chunk_store, n_processes)

    def process(self, chunk_req: ChunkRequest):
        # do nothing (already written locally)
        return True


class GatewayObjStoreOperator(GatewayOperator):
    def __init__(
        self,
        handle: str,
        region: str,
        bucket_name: str,
        bucket_region: str,
        input_queue: GatewayQueue,
        output_queue: GatewayQueue,
        error_event,
        error_queue: Queue,
        n_processes: Optional[int] = 1,
        chunk_store: Optional[ChunkStore] = None,
    ):
        super().__init__(handle, region, input_queue, output_queue, error_event, error_queue, chunk_store, n_processes)
        self.bucket_name = bucket_name
        self.bucket_region = bucket_region
        self.src_requester_pays = cloud_config.get_flag("requester_pays")

        # process-local state
        self.worker_id: Optional[int] = None
        self.obj_store_interfaces: Dict[str, ObjectStoreInterface] = {}

    # interact with object store
    def get_obj_store_interface(self, region: str, bucket: str) -> ObjectStoreInterface:
        key = f"{region}:{bucket}"
        if key not in self.obj_store_interfaces:
            logger.warning(f"[gateway_daemon] ObjectStoreInterface not cached for {key}")
            try:
                self.obj_store_interfaces[key] = ObjectStoreInterface.create(region, bucket)
            except Exception as e:
                raise ValueError(f"Failed to create obj store interface {str(e)}")
        return self.obj_store_interfaces[key]


class GatewayObjStoreReadOperator(GatewayObjStoreOperator):
    def __init__(
        self,
        handle: str,
        region: str,
        bucket_name: str,
        bucket_region: str,
        input_queue: GatewayQueue,
        output_queue: GatewayQueue,
        error_event,
        error_queue: Queue,
        n_processes: int = 32,
        chunk_store: Optional[ChunkStore] = None,
    ):
        super().__init__(
            handle, region, bucket_name, bucket_region, input_queue, output_queue, error_event, error_queue, n_processes, chunk_store
        )

    def process(self, chunk_req: ChunkRequest, **args):
        fpath = str(self.chunk_store.get_chunk_file_path(chunk_req.chunk.chunk_id).absolute())
        # wait for free space
        logger.debug(f"[{self.handle}:{self.worker_id}] Start download {chunk_req.chunk.chunk_id} from {self.bucket_name}")

        obj_store_interface = self.get_obj_store_interface(self.bucket_region, self.bucket_name)

        if self.src_requester_pays:
            obj_store_interface.set_requester_bool(True)

        # while self.chunk_store.remaining_bytes() < chunk_req.chunk.chunk_length_bytes * self.n_processes:
        #    time.sleep(0.1)

        # assert chunk_req.chunk.chunk_length_bytes > 0, f"Cannot have size 0 chunk {chunk_req.chunk}" # actually ok

        if chunk_req.chunk.chunk_length_bytes == 0:
            # nothing to do
            # create empty file
            Path(fpath).touch()
            return True

        while True:
            # if self.chunk_store.remaining_bytes() < chunk_req.chunk.chunk_length_bytes * self.n_processes:
            #    time.sleep(0.1)
            #    continue
            try:
                md5sum = retry_backoff(
                    partial(
                        obj_store_interface.download_object,
                        chunk_req.chunk.src_key,
                        fpath,
                        chunk_req.chunk.file_offset_bytes,
                        chunk_req.chunk.chunk_length_bytes,
                        generate_md5=True,
                    ),
                    max_retries=1,  # TODO: fix this - not a good solution
                )

                # ensure properly downloaded
                file_size = os.path.getsize(fpath)
                if file_size == chunk_req.chunk.chunk_length_bytes:
                    break

            except Exception as e:
                logger.error(f"[obj_store:{self.worker_id}] Error reading key {chunk_req.chunk.src_key}: {str(e)}")
                print(f"[obj_store:{self.worker_id}] Error reading key {chunk_req.chunk.src_key}: {str(e)}")
                time.sleep(1)

        # update md5sum for chunk requests
        # TODO: create checksum operator
        # if not md5sum:
        #    logger.error(f"[obj_store:{self.worker_id}] Checksum was not generated for {chunk_req.chunk.src_key}")
        # else:
        #    self.chunk_store.update_chunk_checksum(chunk_req.chunk.chunk_id, md5sum)

        received_chunk_size = self.chunk_store.get_chunk_file_path(chunk_req.chunk.chunk_id).stat().st_size
        assert (
            received_chunk_size == chunk_req.chunk.chunk_length_bytes
        ), f"Downloaded chunk {chunk_req.chunk.chunk_id} to {fpath} has incorrect size (expected {chunk_req.chunk.chunk_length_bytes} but got {received_chunk_size}, {chunk_req.chunk.chunk_length_bytes})"
        logger.debug(f"[obj_store:{self.worker_id}] Downloaded {chunk_req.chunk.chunk_id} from {self.bucket_name}")
        return True


class GatewayObjStoreWriteOperator(GatewayObjStoreOperator):
    def __init__(
        self,
        handle: str,
        region: str,
        bucket_name: str,
        bucket_region: str,
        input_queue: GatewayQueue,
        output_queue: GatewayQueue,
        error_event,
        error_queue: Queue,
        upload_id_map: DictProxy,  # map of upload_id mappings from client
        n_processes: Optional[int] = 32,
        chunk_store: Optional[ChunkStore] = None,
        prefix: Optional[str] = "",
    ):
        super().__init__(
            handle, region, bucket_name, bucket_region, input_queue, output_queue, error_event, error_queue, n_processes, chunk_store
        )
        self.chunk_store = chunk_store
        self.upload_id_map = upload_id_map
        self.prefix = prefix

    def process(self, chunk_req: ChunkRequest):
        fpath = str(self.chunk_store.get_chunk_file_path(chunk_req.chunk.chunk_id).absolute())
        logger.debug(
            f"[{self.handle}:{self.worker_id}] Start upload {chunk_req.chunk.chunk_id} to {self.bucket_name}, key {chunk_req.chunk.dest_key}"
        )

        # TODO: cache object store interface
        obj_store_interface = self.get_obj_store_interface(self.bucket_region, self.bucket_name)

        if chunk_req.chunk.multi_part:
            upload_id_key = self.bucket_region + chunk_req.chunk.src_key  # format for upload id key
            assert upload_id_key in self.upload_id_map, f"Upload id for {upload_id_key} not found {self.upload_id_map}"
            upload_id = self.upload_id_map[upload_id_key]
        else:
            upload_id = None

        key = self.prefix + chunk_req.chunk.dest_key
        logger.debug(f"[obj_store:{self.worker_id}] Uploading to id {upload_id} for key {key} bucket {self.bucket_name}")
        retry_backoff(
            partial(
                obj_store_interface.upload_object,
                fpath,
                key,
                chunk_req.chunk.part_number,
                upload_id,
                check_md5=chunk_req.chunk.md5_hash,
            ),
            max_retries=1,
        )
        logger.debug(
            f"[obj_store:{self.worker_id}] Uploaded {chunk_req.chunk.chunk_id} partition {chunk_req.chunk.part_number} to {self.bucket_name}"
        )
        return True



# (C) Copyright Samsung SDS. 2023

#

# Licensed under the Apache License, Version 2.0 (the "License");

# you may not use this file except in compliance with the License.

# You may obtain a copy of the License at

#

#     http://www.apache.org/licenses/LICENSE-2.0

#

# Unless required by applicable law or agreed to in writing, software

# distributed under the License is distributed on an "AS IS" BASIS,

# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

# See the License for the specific language governing permissions and

# limitations under the License.
import time
from typing import List, Optional
import uuid

from skyplane.compute.scp.scp_auth import SCPAuthentication
from skyplane.compute.scp.scp_utils import SCPClient
from skyplane.utils import logger


class SCPNetwork:
    def __init__(self, auth: SCPAuthentication, vpc_name="skyplane", sg_name="skyplane"):
        self.auth = auth
        self.scp_client = SCPClient()

    def get_service_zone_id(self, region: str) -> str:
        zones = SCPAuthentication.get_zones()
        for zone in zones:
            if zone["serviceZoneName"] == region:
                return zone["serviceZoneId"]
        raise ValueError(f"Region {region} not found in SCP")

    def get_service_zoneName(self, zoneId: str) -> str:
        zones = SCPAuthentication.get_zones()
        for zone in zones:
            if zone["serviceZoneId"] == zoneId:
                return zone["serviceZoneName"]
        raise ValueError(f"ZoneId {zoneId} not found in SCP")

    def get_vpc_info(self, vpcId):
        url = f"/vpc/v2/vpcs?vpcId={vpcId}"
        response = self.scp_client._get(url)
        return response[0]["serviceZoneId"], response[0]["vpcName"]

    def list_vpcs(self, service_zone_id, vpc_name):
        url = f"/vpc/v2/vpcs?serviceZoneId={service_zone_id}"

        response = self.scp_client._get(url)
        vpcs = [result for result in response if result["vpcName"] == vpc_name]

        if len(vpcs) == 0:
            return None
        else:
            return vpcs

    def get_vpc_detail(self, vpcId):
        url = f"/vpc/v2/vpcs/{vpcId}"
        response = self.scp_client._getDetail(url)
        return response

    def delete_vpc(self, region, vpcId):
        service_zone_id, vpc_name = self.get_vpc_info(vpcId)
        # delete subnets
        subnets = self.list_subnets(vpcId, subnetId=None)
        for subnet in subnets:
            self.delete_subnet(region, subnet["subnetId"], vpcId)

        # delete igw - firewall first
        igws = self.list_igws(vpcId)
        for igw in igws:
            firewallId = self.get_firewallId(igw["internetGatewayId"])
            self.delete_firewall_rules(region, firewallId, None)
            self.delete_igw(region, igw["internetGatewayId"], vpcId)

        # delete security groups
        sgs = self.list_security_groups(vpcId, securityGroupName=None)
        for sg in sgs:
            self.delete_security_group(region, sg["securityGroupId"], sg["securityGroupName"])

        # delete vpc
        url = f"/vpc/v2/vpcs/{vpcId}"
        try:
            self.scp_client._delete(url)
            delete_completion_condition = lambda: not self.list_vpcs(service_zone_id, vpc_name)
            self.scp_client.wait_for_completion(f"[{region}] Deleting SCP Skyplane VPC", delete_completion_condition)

        except Exception as e:
            logger.error(e)
            raise e
        # return response

    def create_vpc(self, region, service_zone_id, vpc_name):
        url = "/vpc/v3/vpcs"
        req_body = {"serviceZoneId": service_zone_id, "vpcName": vpc_name, "vpcDescription": "Skyplane VPC"}
        vpcId = None
        try:
            response = self.scp_client._post(url, req_body)
            vpcId = response["resourceId"]
            create_completion_condition = lambda: self.get_vpc_detail(vpcId)["vpcState"] == "ACTIVE"
            self.scp_client.wait_for_completion(f"[{region}] Creating SCP Skyplane VPC", create_completion_condition)
        except Exception as e:
            logger.error(e)
            raise e
        return vpcId

    def list_igws(self, vpcId: str):
        url = f"/internet-gateway/v2/internet-gateways"
        response = self.scp_client._get(url)
        igws = [result for result in response if result["vpcId"] == vpcId]
        return igws

    def delete_igw(self, region, igwId: str, vpcId: str):
        url = f"/internet-gateway/v2/internet-gateways/{igwId}"
        try:
            self.scp_client._delete(url)
            delete_completion_condition = lambda: not self.list_igws(vpcId)
            self.scp_client.wait_for_completion(f"[{region}] Deleting SCP Skyplane IGW", delete_completion_condition)
        except Exception as e:
            logger.error(e)
            raise e

    def create_igw(self, region, service_zone_id, vpcId: str):
        url = "/internet-gateway/v2/internet-gateways"
        req_body = {
            "firewallEnabled": True,
            "serviceZoneId": service_zone_id,
            "vpcId": vpcId,
            "InternetGatewayDescription": "Default Internet Gateway for Skyplane VPC",
        }
        try:
            response = self.scp_client._post(url, req_body)
            igwId = response["resourceId"]
            return igwId
        except Exception as e:
            logger.error(e)
            raise e

    def check_created_igw(self, region, igwId: str, vpcId: str):
        create_completion_condition = lambda: self.list_igws(vpcId) and self.list_igws(vpcId)[0].get("internetGatewayState") == "ATTACHED"
        self.scp_client.wait_for_completion(f"[{region}] Creating SCP Skyplane Internet Gateway", create_completion_condition)

    def list_subnets(self, vpcId: str, subnetId: Optional[str] = None):
        if subnetId is None:
            url = f"/subnet/v2/subnets?vpcId={vpcId}"
        else:
            url = f"/subnet/v2/subnets?subnetId={subnetId}"

        response = self.scp_client._get(url)
        # subnet = [result for result in response if result['vpcId'] == vpcId]
        return response

    def create_subnet(self, region, vpc_name, vpcId: str):
        url = "/subnet/v2/subnets"
        req_body = {
            "subnetCidrBlock": "192.168.0.0/24",
            "subnetName": vpc_name + "Sub",
            "subnetType": "PUBLIC",
            "vpcId": vpcId,
            "subnetDescription": "Default Subnet for Skyplane VPC",
        }
        try:
            response = self.scp_client._post(url, req_body)
            subnetId = response["resourceId"]
            return subnetId

        except Exception as e:
            logger.error(e)
            raise e

    def check_created_subnet(self, region, subnetId: str, vpcId: str):
        create_completion_condition = lambda: self.list_subnets(vpcId, subnetId)[0]["subnetState"] == "ACTIVE"
        self.scp_client.wait_for_completion(f"[{region}] Creating SCP Skyplane subnet", create_completion_condition)

    def delete_subnet(self, region, subnetId: str, vpcId: str):
        url = f"/subnet/v2/subnets/{subnetId}"
        try:
            response = self.scp_client._delete(url)
            delete_completion_condition = lambda: not self.list_subnets(vpcId, subnetId)
            self.scp_client.wait_for_completion(f"[{region}] Deleting SCP Skyplane subnet", delete_completion_condition)
        except Exception as e:
            logger.error(e)
            raise e
        return response

    def delete_security_group(self, region, securityGroupId: str, securityGroupName: str):
        url = f"/security-group/v2/security-groups/{securityGroupId}"
        try:
            self.scp_client._delete(url)
            delete_completion_condition = lambda: not self.list_security_groups(None, securityGroupName)
            self.scp_client.wait_for_completion(f"[{region}] Deleting SCP Skyplane Security Group", delete_completion_condition)
        except Exception as e:
            logger.error(e)
            raise e

    def list_security_groups(self, vpcId, securityGroupName: Optional[str] = None):
        if securityGroupName is None:
            url = f"/security-group/v2/security-groups?vpcId={vpcId}"
        else:
            url = f"/security-group/v2/security-groups?securityGroupName={securityGroupName}"
        response = self.scp_client._get(url)
        return response

    def create_security_group(self, region, service_zone_id, vpcId: str):
        url = "/security-group/v3/security-groups"
        req_body = {
            "loggable": False,
            "securityGroupName": "SkyplaneSecuGroup",
            "serviceZoneId": service_zone_id,
            "vpcId": vpcId,
            "securityGroupDescription": "Default Security Group for Skyplane VPC",
        }
        try:
            response = self.scp_client._post(url, req_body)
            securityGroupId = response["resourceId"]
            create_completion_condition = (
                lambda: self.list_security_groups(vpcId, req_body["securityGroupName"])[0]["securityGroupState"] == "ACTIVE"
            )
            self.scp_client.wait_for_completion(f"[{region}] Creating SCP Skyplane Security Group", create_completion_condition)
            return securityGroupId

        except Exception as e:
            logger.error(e)
            raise e

    def find_vpc_id(self, service_zone_id, vpc_name):
        vpc_content = self.list_vpcs(service_zone_id, vpc_name)
        if vpc_content is None:
            return None

        vpc_list = [item["vpcId"] for item in vpc_content]
        return vpc_list[0]

    def find_valid_vpc(self, service_zone_id, vpc_name):
        vpc_content = self.list_vpcs(service_zone_id, vpc_name)
        if vpc_content is None:
            return None

        vpc_list = [item["vpcId"] for item in vpc_content if item["vpcState"] == "ACTIVE"]
        for vpc_id in vpc_list:
            igw_list = self.list_igws(vpc_id)
            igw_list = [igw for igw in igw_list if igw["internetGatewayState"] == "ATTACHED"]
            subnet_list = self.list_subnets(vpc_id)
            subnet_list = [subnet for subnet in subnet_list if subnet["subnetState"] == "ACTIVE" and subnet["subnetType"] == "PUBLIC"]
            if len(igw_list) > 0 and len(subnet_list) > 0:
                return vpc_id

        return None

    def define_vpc_name(self, region: str):
        parts = region.split("-")
        return "Skyplane" + parts[0][0:2] + parts[1] + parts[2]

    def make_vpc(self, region: str):
        service_zone_id = self.get_service_zone_id(region)

        vpc_name = self.define_vpc_name(region)

        # find matching valid VPC
        skyplane_vpc = self.find_valid_vpc(service_zone_id, vpc_name)

        try:
            if skyplane_vpc is None:
                # create VPC
                skyplane_vpc = self.create_vpc(region, service_zone_id, vpc_name)

                # create Subnet
                subnet = self.create_subnet(region, vpc_name, skyplane_vpc)

                # create Internet Gateway - firewall, routing table creation is done by SCP
                igw = self.create_igw(region, service_zone_id, skyplane_vpc)

                # create security group
                sg = self.create_security_group(region, service_zone_id, skyplane_vpc)
                self.create_security_group_in_rule(region, sg)
                self.create_security_group_out_rule(region, sg)

                # create firewall rules
                self.check_created_igw(region, igw, skyplane_vpc)
                self.check_created_subnet(region, subnet, skyplane_vpc)

            return skyplane_vpc
        except Exception as e:
            logger.error(e)
            raise e

    def delete_security_group_rules(self, region, securityGroupId: str, ruleIds):
        url = f"/security-group/v2/security-groups/{securityGroupId}/rules"
        req_body = {"ruleDeletionType": "PARTIAL", "ruleIds": ruleIds}
        try:
            self.scp_client._delete(url, req_body)
            delete_completion_condition = lambda: all(not self.list_security_group_rules(securityGroupId, rule) for rule in ruleIds)
            self.scp_client.wait_for_completion(f"[{region}] Deleting SCP Skyplane Security Group Rule", delete_completion_condition)
        except Exception as e:
            logger.error(e)
            raise e

    def list_security_group_rules(self, securityGroupId: str, ruleId: str):
        url = f"/security-group/v2/security-groups/{securityGroupId}/rules"
        response = self.scp_client._get(url)
        if ruleId is not None:
            result = [result for result in response if result["ruleId"] == ruleId]
        else:  # list all ruleIds
            result = [rule["ruleId"] for rule in response]
        return result

    def create_security_group_in_rule(self, region, securityGroupId: str):
        url = f"/security-group/v2/security-groups/{securityGroupId}/rules"
        req_body = {
            "ruleDirection": "IN",
            "services": [
                {
                    "serviceType": "TCP_ALL",
                }
            ],
            "sourceIpAddresses": ["0.0.0.0/0"],
            "ruleDescription": "Skyplane Security Group In rule(ssh)",
        }
        try:
            response = self.scp_client._post(url, req_body)
            ruleId = response["resourceId"]
            create_completion_condition = lambda: self.list_security_group_rules(securityGroupId, ruleId)[0]["ruleState"] == "ACTIVE"
            self.scp_client.wait_for_completion(f"[{region}] Creating SCP Skyplane Security Group In Rule", create_completion_condition)
            return ruleId

        except Exception as e:
            logger.error(e)
            raise e

    def create_security_group_out_rule(self, region, securityGroupId: str):
        url = f"/security-group/v2/security-groups/{securityGroupId}/rules"
        req_body = {
            "ruleDirection": "OUT",
            "services": [
                {
                    "serviceType": "TCP_ALL",
                }
            ],
            "destinationIpAddresses": ["0.0.0.0/0"],
            "ruleDescription": "Skyplane Security Group Out rule",
        }
        try:
            response = self.scp_client._post(url, req_body)
            ruleId = response["resourceId"]
            create_completion_condition = lambda: self.list_security_group_rules(securityGroupId, ruleId)[0]["ruleState"] == "ACTIVE"
            self.scp_client.wait_for_completion(f"[{region}] Creating SCP Skyplane Security Group Out Rule", create_completion_condition)
            return ruleId
        except Exception as e:
            logger.error(e)
            raise e

    def delete_firewall_rules(self, region, firewallId: str, ruleIds, retry_limit=10):
        url = f"/firewall/v2/firewalls/{firewallId}/rules"

        if ruleIds is None:
            ruleIds = self.list_firewall_rules(firewallId, None)

        req_body = {"ruleDeletionType": "PARTIAL", "ruleIds": ruleIds}
        retries = 0
        while retries < retry_limit:
            try:
                # print(f"try delete_firewall_rules: req_body - [{region}] retries : {retries} \n {req_body}")
                response = self.scp_client._delete(url, req_body)
                # delete_completion_condition = lambda: all(not self.list_firewall_rules(firewallId, rule) for rule in ruleIds)
                # self.scp_client.wait_for_completion(f"[{region}] Deleting SCP Skyplane Firewall Rules", delete_completion_condition)
                return response
            except Exception as e:
                # error code 500 retry
                if "500" in e.args[0]:
                    retries += 1
                    # print(f"Error deleting firewall rules(): {e}")
                    logger.fs.debug(f"Error deleting firewall rules(): {e}")
                    if retries == retry_limit:
                        logger.error(e)
                        raise e
                    time.sleep(5)
                else:
                    logger.error(e)
                    raise e

    def get_firewallId(self, igwId: str):
        url = f"/firewall/v2/firewalls?objectId={igwId}"
        result = self.scp_client._get(url)[0]["firewallId"]
        return result

    def get_firewall_rule_details(self, firewallId: str, ruleId: str):
        url = f"/firewall/v2/firewalls/{firewallId}/rules/{ruleId}"
        response = self.scp_client._getDetail(url)
        return response

    def list_firewall_rules(self, firewallId: str, ruleId=None):
        url = f"/firewall/v2/firewalls/{firewallId}/rules?size=1000"
        response = self.scp_client._get(url)
        if ruleId is None:
            result = [rule["ruleId"] for rule in response]
        else:
            result = [result for result in response if result["ruleId"] == ruleId]
        return result

    def add_firewall_22_rule(self, region, servergroup, virtualServerId, retry_limit=5):
        # check rule exist
        def check_rule_exist(server_id, firewall_id):
            ip = self.get_vs_details(server_id)["ip"]
            rules = self.list_firewall_rules(firewall_id, None)
            for rule in rules:
                rule_details = self.get_firewall_rule_details(firewall_id, rule)
                if rule_details["ruleDirection"] == "IN" and rule_details["ruleDescription"] == "Skyplane Firewall ssh rule":
                    if ip in rule_details["destinationIpAddresses"]:
                        # print(f"Pass - {region}-{servergroup} : Already exists firewall rule")
                        return True

        servers = self.list_server_group_instance(servergroup)
        try:
            private_ips = [self.get_vs_details(server["virtualServerId"])["ip"] for server in servers]
        except Exception as e:
            # print(f"Pass - {region}-{servergroup} : Waiting for all servers to be Editing status")
            return e

        vpcId = self.get_vs_details(virtualServerId)["vpcId"]
        igwId = self.list_igws(vpcId)[0]["internetGatewayId"]
        firewallId = self.get_firewallId(igwId)

        url = f"/firewall/v2/firewalls/{firewallId}/rules"
        req_body = {
            "sourceIpAddresses": ["0.0.0.0/0"],
            "destinationIpAddresses": private_ips,
            "services": [{"serviceType": "TCP", "serviceValue": "22"}],
            "ruleDirection": "IN",
            "ruleAction": "ALLOW",
            "isRuleEnabled": True,
            "ruleLocationType": "LAST",
            "ruleDescription": "Skyplane Firewall ssh rule",
        }
        retries = 0
        while retries < retry_limit:
            try:
                if not check_rule_exist(virtualServerId, firewallId):
                    response = self.scp_client.post(url, req_body)
                    ruleId = response["resourceId"]
                    create_completion_condition = lambda: self.get_firewall_rule_details(firewallId, ruleId)["ruleState"] == "ACTIVE"
                    self.scp_client.wait_for_completion(
                        f"[{region}:{private_ips}] Creating SCP Skyplane Firewall 22 Rule", create_completion_condition
                    )
                    return ruleId
                else:
                    return None
            except Exception as e:
                if "500" in e.args[0]:
                    # print(f"Pass - Error creating firewall in rule(): {e}")
                    logger.fs.debug(f"Pass - Error creating firewall in rule(): {e}")
                    return None
                else:
                    retries += 1
                    # print(f"Error creating firewall in rule(): {e}")
                    logger.fs.debug(f"Error creating firewall in rule(): {e}")
                    if retries == retry_limit:
                        logger.fs.error(e)
                        raise e
            time.sleep(1)

    def add_firewall_rule(self, region: str, virtualServerIds: Optional[List[str]] = None):
        for virtualServerId in virtualServerIds:
            try:
                vpcId = self.get_vs_details(virtualServerId)["vpcId"]
                ip = self.get_vs_details(virtualServerId)["ip"]

                igwId = self.list_igws(vpcId)[0]["internetGatewayId"]
                firewallId = self.get_firewallId(igwId)

                self.create_firewall_in_rule(region, firewallId, ip)
                self.create_firewall_out_rule(region, firewallId, ip)
            except Exception as e:
                logger.error(e)
                raise e

    def add_firewall_rule_all(self, region, private_ips, vpcids):
        vpcId = vpcids[0]
        try:
            # igwId = self.list_igws(vpcId)[0]['internetGatewayId']
            igwId = self.get_igw(vpcId)
            firewallId = self.get_firewallId(igwId)

            self.create_firewall_in_rule(region, firewallId, private_ips)
            self.create_firewall_out_rule(region, firewallId, private_ips)
        except Exception as e:
            logger.error(e)
            raise e

    def get_igw(self, vpc_Id: str):
        return self.list_igws(vpc_Id)[0]["internetGatewayId"]

    def create_firewall_in_rule(self, region, firewallId, internalIp, retry_limit=60):
        url = f"/firewall/v2/firewalls/{firewallId}/rules"
        req_body = {
            "sourceIpAddresses": ["0.0.0.0/0"],
            "destinationIpAddresses": internalIp,
            "services": [{"serviceType": "TCP_ALL"}],
            "ruleDirection": "IN",
            "ruleAction": "ALLOW",
            "isRuleEnabled": True,
            "ruleLocationType": "LAST",
            "ruleDescription": "Skyplane Firewall In rule",
        }
        retries = 0
        while retries < retry_limit:
            try:
                response = self.scp_client._post(url, req_body)
                # print(response)
                # time.sleep(3)
                ruleId = response["resourceId"]
                create_completion_condition = lambda: self.get_firewall_rule_details(firewallId, ruleId)["ruleState"] == "ACTIVE"
                self.scp_client.wait_for_completion(f"[{region}] Creating SCP Skyplane Firewall In Rule", create_completion_condition)
                return ruleId
            except Exception as e:
                retries += 1
                print(f"Error creating firewall in rule(): {e}")
                if retries == retry_limit:
                    logger.error(e)
                    raise e
            time.sleep(2)

    def create_firewall_out_rule(self, region, firewallId, internalIp, retry_limit=60):
        url = f"/firewall/v2/firewalls/{firewallId}/rules"
        req_body = {
            "sourceIpAddresses": internalIp,
            "destinationIpAddresses": ["0.0.0.0/0"],
            "services": [{"serviceType": "TCP_ALL"}],
            "ruleDirection": "OUT",
            "ruleAction": "ALLOW",
            "isRuleEnabled": True,
            "ruleLocationType": "FIRST",
            "ruleDescription": "Skyplane Firewall Out rule",
        }
        retries = 0
        while retries < retry_limit:
            try:
                response = self.scp_client._post(url, req_body)
                ruleId = response["resourceId"]
                create_completion_condition = lambda: self.get_firewall_rule_details(firewallId, ruleId)["ruleState"] == "ACTIVE"
                self.scp_client.wait_for_completion(f"[{region}] Creating SCP Skyplane Firewall Out Rule", create_completion_condition)
                return ruleId
            except Exception as e:
                retries += 1
                print(f"Error creating firewall out rule(): {e}")
                if retries == retry_limit:
                    logger.error(e)
                    raise e
            time.sleep(2)

    def list_instances(self, instanceId):
        if instanceId is None:
            url = f"/virtual-server/v2/virtual-servers"
        else:
            try:
                instanceName = self.get_vs_details(instanceId)["virtualServerName"]
            except Exception as e:
                if "404" in e.args[0]:
                    return None
            url = f"/virtual-server/v2/virtual-servers?virtualServerName={instanceName}"

        result = self.scp_client._get(url)
        if len(result) == 0:
            return None
        else:
            return result

    def start_instance(self, instanceId: str):
        url = f"/virtual-server/v2/virtual-servers/{instanceId}/start"
        try:
            response = self.scp_client._post(url, req_body={})
            instanceId = response["resourceId"]
            completion_condition = lambda: self.list_instances(instanceId)[0]["virtualServerState"] == "RUNNING"
            self.scp_client.wait_for_completion(f" Starting SCP Skyplane Virtual Server", completion_condition)
        except Exception as e:
            logger.error(e)
            raise e

    def stop_instance(self, instanceId: str):
        url = f"/virtual-server/v2/virtual-servers/{instanceId}/stop"
        try:
            response = self.scp_client._post(url, req_body={})
            instanceId = response["resourceId"]
            completion_condition = lambda: self.list_instances(instanceId)[0]["virtualServerState"] == "STOPPED"
            self.scp_client.wait_for_completion(f" Stopping SCP Skyplane Virtual Server", completion_condition)
        except Exception as e:
            logger.error(e)
            raise e

    def terminate_instance(self, instanceId: str):
        url = f"/virtual-server/v2/virtual-servers/{instanceId}"
        try:
            response = self.scp_client._delete(url, req_body={})
            instanceId = response["resourceId"]
            # completion_condition = lambda: self.list_instances(instanceId) is None
            # self.scp_client.wait_for_completion(f" Terminating SCP Skyplane Virtual Server", completion_condition)
        except Exception as e:
            logger.error(e)
            raise e

    def get_nic_details(self, virtualServerId):
        url = f"/virtual-server/v2/virtual-servers/{virtualServerId}/nics"
        response = self.scp_client._get(url)
        return response

    def get_vs_details(self, virtualServerId):
        url = f"/virtual-server/v3/virtual-servers/{virtualServerId}"
        response = self.scp_client._getDetail(url)
        return response

    def create_server_group(self, region):
        url = f"/server-group/v2/server-groups"
        serviceZoneId = self.get_service_zone_id(region)
        req_body = {
            "serverGroupName": f"skyplane-{uuid.uuid4().hex[:8]}",
            "serviceZoneId": serviceZoneId,
            "serviceFor": "VIRTUAL_SERVER",
            "servicedGroupFor": "COMPUTE",
        }
        response = self.scp_client._post(url, req_body)
        serverGroupId = response["serverGroupId"]
        return serverGroupId

    def list_server_group(self):
        url = f"/server-group/v2/server-groups"
        response = self.scp_client._get(url)
        return response

    def delete_server_group(self, serverGroupId):
        url = f"/server-group/v2/server-groups/{serverGroupId}"
        if len(self.list_server_group_instance(serverGroupId)) == 0:
            response = self.scp_client._delete(url)
            return response

    def detail_server_group(self, serverGroupId):
        url = f"/server-group/v2/server-groups/{serverGroupId}"
        response = self.scp_client._getDetail(url)
        return response

    def list_server_group_instance(self, serverGroupId):
        url = f"/virtual-server/v2/virtual-servers?serverGroupId={serverGroupId}"
        response = self.scp_client._get(url)
        return response
